/*
 * Copyright (C) 2011 Baidu Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

//#define LOG_NDEBUG 0 
#define LOG_TAG "FFMPEG_API"
#include <utils/Log.h>
#include <dlfcn.h>

#include "ffmpeg_api.h"
#include "avcodec.h"
#include "assert.h"

#ifdef HAVE_AV_CONFIG_H
#undef HAVE_AV_CONFIG_H
#endif

#include "config.h"
#include "dsputil.h"
#include "libavutil/mathematics.h"
#include "libavutil/imgutils.h"
#include "libavutil/log.h"
#include "libavutil/pixfmt.h"
#include "libavcodec/audioconvert.h"

//#define FFAPI_DEBUG
//#define DUMP_COMPRESSED_DATA
#ifdef DUMP_COMPRESSED_DATA
static FILE *gCompressedFile = NULL;
static char * gCompressedPath = "/data/dumped_ffmpeg_data.ffmpg";
#endif

typedef int (*GetYUVBufferFun)(void *handle, void **pBase);
typedef int (*ReleaseYUVBufferFun)(void *handle, void *pBase);


typedef struct{
/* global codec objects */
    AVCodec        *gpAVCodec ;
    AVCodecContext *gpAVCtx;
    AVFrame        *gpAVPicture;
    AVPacket        gAVPacket;
    AVAudioConvert *reformat_ctx;
    int         giCodecID;
    int         giAVGotPic;
    int         gDecOpen;
    int         avflag;
    int         userBufferFlag;

    void *bufManHandle;
    GetYUVBufferFun getYUVBuffer;
    ReleaseYUVBufferFun releaseYUVBuffer;
}FFCODEC;

enum {
    FF_AUDIO ,
    FF_VIDEO
};


static void avlogAndroidCallback(void* ptr, int level, const char* fmt, va_list vl){
#ifdef FFAPI_DEBUG
    char line[512];
    vsnprintf(line, sizeof(line) - 1, fmt, vl);
    if(level == AV_LOG_QUIET || level == AV_LOG_PANIC || 
            level == AV_LOG_FATAL || level == AV_LOG_ERROR){
        LOGE("%s", line);
    }else if(level == AV_LOG_WARNING){
        LOGW("%s", line);
    }else if(level == AV_LOG_INFO){
        LOGI("%s", line);
    }else{
        LOGV("%s", line);
    }
#endif
}

static void dumpStreamBuffer(const unsigned char *pucStreamBuffer, int iSize){
    int i;
    unsigned char * data = (unsigned char *)pucStreamBuffer;
    LOGV("pucStreamBuffer, size = %d", iSize);
    iSize = (iSize > 64 ? 64 : iSize);
    for(i = 0; i < iSize; i += 16){
        LOGV("%2x %2x %2x %2x, %2x %2x %2x %2x, %2x %2x %2x %2x, %2x %2x %2x %2x",
            data[i + 0], data[i + 1], data[i + 2], data[i + 3],
            data[i + 4], data[i + 5], data[i + 6], data[i + 7],
            data[i + 8], data[i + 9], data[i + 10], data[i + 11],
            data[i + 12], data[i + 13], data[i + 14], data[i + 15]
        );
    }
}

static int FFCallback_get_buffer(AVCodecContext *s, AVFrame *pic){
    int i;
    int w= s->width;
    int h= s->height;
    int *picture_number;
    uint8_t* pAvailBuf = NULL;
#ifdef FFAPI_DEBUG
    av_log(s, AV_LOG_DEBUG, "Call back to get buffer\n");
#endif
    assert(s->opaque != NULL);

    FFCODEC* ffconfig =(FFCODEC *)(s->opaque);
    if(ffconfig == NULL){
        return -1;
    }

    if(pic->data[0]!=NULL) {
        av_log(s, AV_LOG_ERROR, "pic->data[0]!=NULL in get_buffer()\n");
        return -1;
    }

    if(av_image_check_size(w, h, 0, s)){
        return -1;
    }

    {
        int h_chroma_shift, v_chroma_shift;
        int size[4] = {0};
        int tmpsize;
        int unaligned;
        AVPicture picture;
        int stride_align[4];
        int offset = 0;
        avcodec_get_chroma_sub_sample(s->pix_fmt, &h_chroma_shift, &v_chroma_shift);

        avcodec_align_dimensions2(s, &w, &h, stride_align);

        if(!(s->flags&CODEC_FLAG_EMU_EDGE)){
            w+= EDGE_WIDTH*2;
            h+= EDGE_WIDTH*2;
        }

        do {
            // NOTE: do not align linesizes individually, this breaks e.g. assumptions
            // that linesize[0] == 2*linesize[1] in the MPEG-encoder for 4:2:2
            av_image_fill_linesizes(picture.linesize, s->pix_fmt, w);
            // increase alignment of w for next try (rhs gives the lowest bit set in w)
            w += w & ~(w-1);

            unaligned = 0;
            for (i=0; i<4; i++){
                unaligned |= picture.linesize[i] % stride_align[i];
            }
        } while (unaligned);

        tmpsize = av_image_fill_pointers(picture.data, s->pix_fmt, h, NULL, picture.linesize);
        if (tmpsize < 0)
            return -1;

        for (i=0; i<3 && picture.data[i+1]; i++)
            size[i] = picture.data[i+1] - picture.data[i];
        size[i] = tmpsize - (picture.data[i] - picture.data[0]);

#ifdef FFAPI_DEBUG
        av_log(s, AV_LOG_DEBUG, "size[] = %d, %d, %d, %d,\n", size[0], size[1], size[2], size[3]);
#endif

        if(ffconfig->getYUVBuffer(ffconfig->bufManHandle, (void**)&pAvailBuf) != FF_OK){
            av_log(s, AV_LOG_ERROR, "getYUVBuffer return error!\n");
            return -1;
        }

        for(i=0; i<4 && size[i]; i++){
            const int h_shift= i==0 ? 0 : h_chroma_shift;
            const int v_shift= i==0 ? 0 : v_chroma_shift;

            pic->linesize[i]= picture.linesize[i];

            pic->base[i] = pAvailBuf + offset;
            offset += size[i]+16;

            if (i == 0) {
                offset += size[i] / 2;
            }

            // no edge if EDGE EMU or not planar YUV
            if((s->flags&CODEC_FLAG_EMU_EDGE) || !size[2])
                pic->data[i] = pic->base[i];
            else
                pic->data[i] = pic->base[i] + FFALIGN((pic->linesize[i]*EDGE_WIDTH>>v_shift) + (EDGE_WIDTH>>h_shift), stride_align[i]);
        }
        if(size[1] && !size[2])
            ff_set_systematic_pal2((uint32_t*)pic->data[1], s->pix_fmt);

        pic->age= INT_MAX;

#ifdef FFAPI_DEBUG
        av_log(s, AV_LOG_DEBUG, "s->width = %d, s->height = %d\n",
            s->width, s->height);
        av_log(s, AV_LOG_DEBUG, "stride_align[0] = %d, stride_align[1] = %d, stride_align[2] = %d\n", 
            stride_align[0], stride_align[1], stride_align[2]);
        av_log(s, AV_LOG_DEBUG, "linesize[0] = %d, linesize[1] = %d, linesize[2] = %d, h = %d\n",
            pic->linesize[0], pic->linesize[1], pic->linesize[2], h);
        av_log(s, AV_LOG_DEBUG, "pic->base[0] = 0x%x, pic->base[1] = 0x%x, pic->base[2] = 0x%x\n",
            (uint)pic->base[0], (uint)pic->base[1], (uint)pic->base[2]);
#endif
    }
    pic->type= FF_BUFFER_TYPE_USER;

    if(s->pkt) pic->pkt_pts= s->pkt->pts;
    else       pic->pkt_pts= AV_NOPTS_VALUE;
    pic->reordered_opaque= s->reordered_opaque;

    av_log(s, AV_LOG_VERBOSE, "get buffer return OK\n");
    return 0;
}

static void FFCallback_release_buffer(AVCodecContext *s, AVFrame *pic){
    int i;
    FFCODEC* ffconfig =(FFCODEC *)(s->opaque);
#ifdef FFAPI_DEBUG
    av_log(s, AV_LOG_DEBUG, "Call back to release buffer\n");
#endif
    if(pic == NULL || pic->type != FF_BUFFER_TYPE_USER || pic->base[0] == NULL){
        LOGW("Attempt to release non exist buffer!");
        LOGI("pic = 0x%x, pic->base[0] = 0x%x", (uint)pic, pic == NULL ? 0 : (uint)pic->base[0]);
        return;
    }

    if(ffconfig->releaseYUVBuffer(ffconfig->bufManHandle, pic->base[0]) != FF_OK){
        av_log(s, AV_LOG_ERROR, "releaseYUVBuffer return error!\n");
    }

    for(i=0; i<4; i++){
        pic->data[i]=NULL;
        pic->base[i]=NULL;
    }

}


static void wrapCallbackFunctions(AVCodecContext * avctx, int codecID){
    avctx->get_buffer= FFCallback_get_buffer;
    avctx->release_buffer= FFCallback_release_buffer;
}

static void setMetaDataParam(FFCODEC *ffconfig, MetaConfig_t *metaconfig){
    if((ffconfig->giCodecID >= CODEC_ID_PCM_S16LE)&&(CODEC_ID_CELT >= ffconfig->giCodecID))
    {   /* Audio codec */
        ffconfig->gpAVCtx->channels = metaconfig->channels;
        if(ffconfig->gpAVCtx->channels > 2) {
            ffconfig->gpAVCtx->request_channels = 2;
        }
        ffconfig->gpAVCtx->sample_rate = metaconfig->sample_rate;
        ffconfig->gpAVCtx->bit_rate = metaconfig->bit_rate;
        ffconfig->gpAVCtx->block_align = metaconfig->block_align;   //TODO   
        ffconfig->avflag=FF_AUDIO;        
    }
    else if((ffconfig->giCodecID >= CODEC_ID_MPEG1VIDEO)&&(CODEC_ID_DFA >= ffconfig->giCodecID))
    {   /* Video codec */
        ffconfig->gpAVCtx->width    = metaconfig->width;
        ffconfig->gpAVCtx->height   = metaconfig->height;
        ffconfig->avflag=FF_VIDEO; 
    }
    else
    {
        LOGE("codec ID = 0x%x is unknown", ffconfig->giCodecID);
    }
    ffconfig->gpAVCtx->bits_per_coded_sample = metaconfig->bits_per_sample;
    return;
} 


void* FFAPI_InitCodec(MetaConfig_t *metaconfig, int iCodecID,  BufManConfig_t *bufManConfig)
{
    FFCODEC *ffconfig = NULL;
    LOGV("metaconfig = 0x%x", (uint32_t)metaconfig);
    LOGV("bufManConfig = 0x%x", (uint32_t)bufManConfig);    
    ffconfig = malloc(sizeof(FFCODEC));
    if(ffconfig == NULL){
        LOGE("ffconfig == NULL");
        return NULL;
    }
    memset(ffconfig, 0, sizeof(FFCODEC));
    ffconfig->gpAVCodec = NULL;
    ffconfig->gpAVCtx = NULL;
    ffconfig->reformat_ctx = NULL;
    ffconfig->gpAVPicture = NULL;
    ffconfig->giCodecID = 0x88888888;
    ffconfig->giAVGotPic  = 0;
    ffconfig->gDecOpen  = 0;
    ffconfig->userBufferFlag = 0;

    av_log_set_callback(avlogAndroidCallback);

    av_init_packet(&ffconfig->gAVPacket);

    /* must be called before using avcodec lib */
    avcodec_init();

    /* register all the codecs */
    avcodec_register_all();

    /* get codec contents */
    ffconfig->giCodecID = iCodecID;
    

    if(!(ffconfig->gpAVCodec = avcodec_find_decoder(iCodecID))) {
        LOGE("Find decoder fail, CodecID is %d", iCodecID);
        free(ffconfig);
        return NULL;
    }

    ffconfig->gpAVCtx = avcodec_alloc_context();
    if(ffconfig->gpAVCtx==NULL){
        LOGE("Allocate context fail, Codec name is %s", ffconfig->gpAVCodec->name);
        free(ffconfig);
        return NULL;
    }

    setMetaDataParam(ffconfig, metaconfig);

    ffconfig->gpAVPicture = avcodec_alloc_frame();
    if(ffconfig->gpAVPicture == NULL){
        LOGE("Allocate frame fail, codec name is %s", ffconfig->gpAVCodec->name);
        avcodec_close(ffconfig->gpAVCtx);
        av_free(ffconfig->gpAVCtx);
        ffconfig->gpAVCtx = NULL;
        free(ffconfig);
        return NULL;
    }

    if(bufManConfig && bufManConfig->user_buffer_flag == 1){
        ffconfig->gpAVCtx->opaque = (void*)ffconfig;
        wrapCallbackFunctions(ffconfig->gpAVCtx, iCodecID);
        ffconfig->userBufferFlag = 1;
        ffconfig->bufManHandle = bufManConfig->buf_man_handle;
        if(bufManConfig->omx_flag == 1){
            void* LibHandle = dlopen("libBaiduOmxFFmpegVdec.so", RTLD_NOW);
            ffconfig->getYUVBuffer = (GetYUVBufferFun)dlsym(LibHandle, "Baidu_OMX_GetYUVBuffer");
            ffconfig->releaseYUVBuffer = (ReleaseYUVBufferFun)dlsym(LibHandle, "Baidu_OMX_ReleaseYUVBuffer");
            if (LibHandle) {
                dlclose(LibHandle);
                LibHandle = NULL;
            }
        }else{
            void* LibHandle = dlopen("libstagefright.so", RTLD_NOW);
            ffconfig->getYUVBuffer = (GetYUVBufferFun)dlsym(LibHandle, "GetYUVBuffer");
            ffconfig->releaseYUVBuffer = (ReleaseYUVBufferFun)dlsym(LibHandle, "ReleaseYUVBuffer");
            if (LibHandle) {
                dlclose(LibHandle);
                LibHandle = NULL;
            }
        }
        LOGV("getYUVBuffer = 0x%x, releaseYUVBuffer = 0x%x", ffconfig->getYUVBuffer, ffconfig->releaseYUVBuffer);
    }

/*
    if(ffconfig->gpAVCodec->capabilities&CODEC_CAP_TRUNCATED)
        ffconfig->gpAVCtx->flags|= CODEC_FLAG_TRUNCATED;*/ /* we do not send complete frames */


#ifdef DUMP_COMPRESSED_DATA
    if(ffconfig->avflag == FF_VIDEO){
        LOGV("FFAPI_InitCodec, codecID = %d, userBufferFlag = %d", iCodecID, bufManConfig->user_buffer_flag);
        LOGV("metaconfig->width = %d, metaconfig->height = %d", metaconfig->width, metaconfig->height);

        if(gCompressedFile== NULL){
            //gCompressedFile = fopen(gCompressedPath, "wb"); 
            
            if(gCompressedFile == NULL){
               LOGE("FAIL TO  OPEN THE DUMP FILE:  %s", gCompressedPath);
            }
        }
    }
#endif 

    /* For some codecs, such as msmpeg4 and mpeg4, width and height
       MUST be initialized there because this information is not
       available in the bitstream. */
    LOGV("init ffmpeg codec ok");
    return ffconfig;
}

int FFAPI_SetExtraData(void* handle,MetaConfig_t *pucExtraData)
{
    FFCODEC *ffconfig=(FFCODEC *)handle;	
    MetaConfig_t *metaconfig=(MetaConfig_t *)pucExtraData;

#ifdef DUMP_COMPRESSED_DATA
        if(ffconfig->avflag == FF_VIDEO){
            LOGV("FFAPI_SetExtraData, metaconfig->extradata_size = %d", metaconfig->extradata_size);
            if(gCompressedFile){
                if (metaconfig->extradata_size > 0) {
                    fwrite(metaconfig->extradata , 1, metaconfig->extradata_size, gCompressedFile);
                }
            }
        }
#endif 

    if (!handle || !ffconfig->gpAVCtx || !ffconfig->gpAVCodec) {
        LOGE ("codec hasn't been initialized.\n");
        return FF_FAIL;
    }

    if (metaconfig->extradata_size > 0) {
        ffconfig->gpAVCtx->extradata_size = metaconfig->extradata_size;
        ffconfig->gpAVCtx->extradata = (uint8_t *)malloc(metaconfig->extradata_size);
        memcpy(ffconfig->gpAVCtx->extradata, metaconfig->extradata, metaconfig->extradata_size);
        if (((ffconfig->giCodecID == CODEC_ID_RAWVIDEO) ||
            (ffconfig->giCodecID == CODEC_ID_MSVIDEO1)) &&
            (metaconfig->extradata_size >= AVPALETTE_COUNT * 4)) {
             int k;
             uint32_t *palette = (uint32_t *)ffconfig->gpAVCtx->extradata;
             if (ffconfig->gpAVCtx->palctrl == NULL) {
                   ffconfig->gpAVCtx->palctrl = malloc(sizeof(AVPaletteControl));
             }
             for (k = 0; k < AVPALETTE_COUNT; k++) {
                       ffconfig->gpAVCtx->palctrl->palette[k] = palette[k];
             }
        }        
    }
    return FF_OK;
}


int FFAPI_Decode(void* handle,const unsigned char * pucStreamBuffer, 
        int iSize, int * piBytesConsumed, int *videoDecodeEnd,
        int *piOutsize,short* pOutputBuffer){

    FFCODEC *ffconfig=(FFCODEC *)handle;

#ifdef DUMP_COMPRESSED_DATA
    static int call_cnt = 0;
    if(ffconfig->avflag == FF_VIDEO){
        LOGV("FFAPI_Decode, call cnt = %d, iSize = %d", call_cnt++, iSize);
        if(gCompressedFile){
            fwrite(pucStreamBuffer , 1, iSize, gCompressedFile);
        }
    }
#endif

    /* since some codec needs extra data, we can't open codec before extra data is set */
    if (ffconfig->gDecOpen == 0) {
        /* open it */
        if (avcodec_open(ffconfig->gpAVCtx, ffconfig->gpAVCodec) < 0) {            
            return FF_FAIL;
        }
        else{
            ffconfig->gDecOpen = 1;
        }
    }

    /* decode */
    int len = 0;


    ffconfig->gAVPacket.data = (unsigned char *)pucStreamBuffer;
    ffconfig->gAVPacket.size = iSize;


    if(ffconfig->avflag == FF_AUDIO){  
        
        if ((len = avcodec_decode_audio3(ffconfig->gpAVCtx,
            pOutputBuffer,  piOutsize, &ffconfig->gAVPacket)) < 0) {
            *piBytesConsumed = iSize;            
            return FF_FAIL;
        }
        if(ffconfig->gpAVCtx->sample_fmt == AV_SAMPLE_FMT_FLT){
            if(ffconfig->reformat_ctx == NULL)
                ffconfig->reformat_ctx = av_audio_convert_alloc(AV_SAMPLE_FMT_S16, 1,
                                                       AV_SAMPLE_FMT_FLT, 1, NULL, 0);
            const void *ibuf[6]= {pOutputBuffer};
            void *obuf[6]= {pOutputBuffer};
            int istride[6]= {4};
            int ostride[6]= {2};
            int len= *piOutsize/istride[0];
            av_audio_convert(ffconfig->reformat_ctx, obuf, ostride, ibuf, istride, len);
            *piOutsize = *piOutsize/2;
        }
        if(ffconfig->gpAVCtx->channels == 6){
            int len = *piOutsize;
            int i;
            for(i = 0; i < len/12; i++){
                *(pOutputBuffer + i*2) = (*(pOutputBuffer + i*6) + *(pOutputBuffer + i*6 + 2)) / 2;
                *(pOutputBuffer + i *2+ 1) = (*(pOutputBuffer + i*6 + 1) + *(pOutputBuffer + i*6 + 2)) / 2;  

            }

            *piOutsize = *piOutsize/3;
            

        }
    }else{
        if ((len = avcodec_decode_video2(ffconfig->gpAVCtx,
        ffconfig->gpAVPicture, &ffconfig->giAVGotPic, &ffconfig->gAVPacket)) < 0) {
            /* error while decoding frame */
            *videoDecodeEnd = (ffconfig->giAVGotPic || (ffconfig->gAVPacket.size <= 0));
            *piBytesConsumed = iSize;
            return FF_FAIL;
        }
    }

    ffconfig->gAVPacket.size -= len;
    ffconfig->gAVPacket.data += len;

    if(ffconfig->avflag == FF_VIDEO){
        *videoDecodeEnd = (ffconfig->giAVGotPic || (ffconfig->gAVPacket.size <= 0));
    }

    *piBytesConsumed = len; 
    return FF_OK;
    
}


int FFAPI_GetFrameBuffer(void* handle,FFLIBFRMBUF *pFrame)
{
    FFCODEC *ffconfig=(FFCODEC *)handle;
    if (ffconfig->giAVGotPic && ffconfig->gpAVPicture) {
        /* we pass the frame buffer information to upper layer */
        pFrame->base = ffconfig->gpAVPicture->base[0];
        pFrame->yuv[0] = ffconfig->gpAVPicture->data[0];
        pFrame->yuv[1] = ffconfig->gpAVPicture->data[1];
        pFrame->yuv[2] = ffconfig->gpAVPicture->data[2];
        pFrame->pitch[0] = ffconfig->gpAVPicture->linesize[0];
        pFrame->pitch[1] = ffconfig->gpAVPicture->linesize[1];
        pFrame->pitch[2] = ffconfig->gpAVPicture->linesize[2];
        pFrame->pic_width  = ffconfig->gpAVCtx->width;
        pFrame->pic_height = ffconfig->gpAVCtx->height;
        pFrame->pixel_format = ffconfig->gpAVCtx->pix_fmt;
        LOGV("pFrame->pixel_format = %d", pFrame->pixel_format);
        if (ffconfig->giCodecID == CODEC_ID_RAWVIDEO) {
//now pFrame->yuv[0] is equal packet data, size[0] = linesizes[0] * height;           
              if (ffconfig->gpAVPicture->data[0]) {
                   FFCallback_release_buffer(ffconfig->gpAVCtx,ffconfig->gpAVPicture);
              }
              ffconfig->gpAVPicture->data[0] = 0;//first is not zero
              ffconfig->gpAVPicture->data[1] = 0;
              if (FFCallback_get_buffer(ffconfig->gpAVCtx,ffconfig->gpAVPicture) < 0) {
                   return FF_FAIL;
              }
//             LOGV("RAWVideo base 0x%x\n",ffconfig->gpAVPicture->base[0]);
//             LOGV("RAWVideo yuv[0] 0x%x,yuv[1] 0x%x,yuv[2] 0x%x,pitch[0] %d,pitch[1]  %d,pitch[2] %d\n",
//                  ffconfig->gpAVPicture->data[0],ffconfig->gpAVPicture->data[1],ffconfig->gpAVPicture->data[2] ,
//                  pFrame->pitch[0],pFrame->pitch[1],pFrame->pitch[2]);
//it same as bitmap invert data,maybe all of BMP are so?
             
              if (pFrame->pixel_format == PIX_FMT_PAL8) {
                    int i,j;
                    unsigned char *dst = ffconfig->gpAVPicture->data[0] + pFrame->pitch[0] * (pFrame->pic_height -1) ;
                    unsigned char *src = pFrame->yuv[0];
                    for (j = 0; j < pFrame->pic_height; j++) {
                             for (i = 0; i < pFrame->pic_width; i++) {
                                      dst[i] = src[i];
                             }
                             src += pFrame->pitch[0];
                             dst -= pFrame->pitch[0];
                    }
              }else {
                    memcpy(ffconfig->gpAVPicture->data[0],pFrame->yuv[0], pFrame->pitch[0] * pFrame->pic_height);
              }
              memcpy(ffconfig->gpAVPicture->data[1],pFrame->yuv[1],AVPALETTE_SIZE);              
              pFrame->base = ffconfig->gpAVPicture->base[0];
              pFrame->yuv[0] = ffconfig->gpAVPicture->data[0];
              pFrame->yuv[1] = ffconfig->gpAVPicture->data[1];
        }
        return FF_OK;
    }

    return FF_FAIL;
}

void FFAPI_Reset(void* handle)
{
    FFCODEC *ffconfig = (FFCODEC *)handle;
    if (ffconfig->gpAVCtx && ffconfig->gpAVCodec) {
        /* close then re-open */
        avcodec_close(ffconfig->gpAVCtx);
        ffconfig->gDecOpen = 0;
        avcodec_open(ffconfig->gpAVCtx, ffconfig->gpAVCodec);
    }
} 

void FFAPI_FlushBuffers(void* handle)
{
    if (handle != NULL) {
        FFCODEC *ffconfig = (FFCODEC *)handle;
        if(ffconfig->gpAVCtx && ffconfig->gpAVCtx->codec){
            avcodec_flush_buffers(ffconfig->gpAVCtx);
        }
    }
}

void FFAPI_FreeCodec(void* handle)
{
    if (handle != NULL) {
        FFCODEC *ffconfig = (FFCODEC *)handle;

        if (ffconfig->gpAVCtx->extradata) {
            free(ffconfig->gpAVCtx->extradata);
            ffconfig->gpAVCtx->extradata = NULL;
        }
        if (ffconfig->gpAVCtx->palctrl) {
              free(ffconfig->gpAVCtx->palctrl);
              ffconfig->gpAVCtx->palctrl = NULL;
        }
        if (ffconfig->gpAVCtx) {
            avcodec_close(ffconfig->gpAVCtx);
            av_free(ffconfig->gpAVCtx);
            ffconfig->gpAVCtx = NULL;
        }
        if (ffconfig->reformat_ctx)
            av_audio_convert_free(ffconfig->reformat_ctx);

        if (ffconfig->gpAVPicture) {
            av_free(ffconfig->gpAVPicture);
            ffconfig->gpAVPicture = NULL;
        }
        ffconfig->gDecOpen = 0;
        free(ffconfig);
    }
} 


void FFAPI_DisplayFrameInfo(const FFLIBFRMBUF* pFrame){
    LOGD("YUV buffer[0]:0x%x, [1]:0x%x, [2]:0x%x\n",(uint)pFrame->yuv[0], (uint)pFrame->yuv[1], (uint)pFrame->yuv[2]);
}


